@@ -8,6 +8,11 @@
 //! This crate **does not require** Python 3 distribution files
 //! to be present on the (cross-)compile host system.
 //!
+//! This crate uses the binutils `dlltool` program to generate
+//! the Python DLL import libraries for MinGW-w64 targets.
+//! Setting `PYO3_MINGW_DLLTOOL` environment variable overrides
+//! the default `dlltool` command name for the target.
+//!
 //! **Note:** MSVC cross-compile targets require either LLVM binutils
 //! or Zig to be available on the host system.
 //! More specifically, `python3-dll-a` requires `llvm-dlltool` executable
@@ -118,6 +123,15 @@ const DLLTOOL_MSVC: &str = "llvm-dlltool";
 #[cfg(windows)]
 const LIB_MSVC: &str = "lib.exe";
 
+/// Python interpreter implementations
+#[derive(Debug, Clone, Copy)]
+pub enum PythonImplementation {
+    /// CPython
+    CPython,
+    /// PyPy
+    PyPy,
+}
+
 /// Windows import library generator for Python
 ///
 /// Generates `python3.dll` or `pythonXY.dll` import library directly from the
@@ -159,6 +173,8 @@ pub struct ImportLibraryGenerator {
     env: String,
     /// Major and minor Python version (for `pythonXY.dll` only)
     version: Option<(u8, u8)>,
+    /// Python interpreter implementation
+    implementation: PythonImplementation,
 }
 
 impl ImportLibraryGenerator {
@@ -174,6 +190,7 @@ impl ImportLibraryGenerator {
             arch: arch.to_string(),
             env: env.to_string(),
             version: None,
+            implementation: PythonImplementation::CPython,
         }
     }
 
@@ -185,6 +202,12 @@ impl ImportLibraryGenerator {
         self
     }
 
+    /// Sets Python interpreter implementation
+    pub fn implementation(&mut self, implementation: PythonImplementation) -> &mut Self {
+        self.implementation = implementation;
+        self
+    }
+
     /// Generates the Python DLL import library in `out_dir`.
     ///
     /// The version-agnostic `python3.dll` import library is generated
@@ -224,14 +247,22 @@ impl ImportLibraryGenerator {
     ///
     /// Returns the newly created `python3.def` or `pythonXY.def` file path.
     fn write_def_file(&self, out_dir: &Path) -> Result<PathBuf> {
-        let (def_file, def_file_content) = match self.version {
-            None => ("python3.def", include_str!("python3.def")),
-            Some((3, 7)) => ("python37.def", include_str!("python37.def")),
-            Some((3, 8)) => ("python38.def", include_str!("python38.def")),
-            Some((3, 9)) => ("python39.def", include_str!("python39.def")),
-            Some((3, 10)) => ("python310.def", include_str!("python310.def")),
-            Some((3, 11)) => ("python311.def", include_str!("python311.def")),
-            _ => return Err(Error::new(ErrorKind::Other, "Unsupported Python version")),
+        let (def_file, def_file_content) = match self.implementation {
+            PythonImplementation::CPython => match self.version {
+                None => ("python3.def", include_str!("python3.def")),
+                Some((3, 7)) => ("python37.def", include_str!("python37.def")),
+                Some((3, 8)) => ("python38.def", include_str!("python38.def")),
+                Some((3, 9)) => ("python39.def", include_str!("python39.def")),
+                Some((3, 10)) => ("python310.def", include_str!("python310.def")),
+                Some((3, 11)) => ("python311.def", include_str!("python311.def")),
+                _ => return Err(Error::new(ErrorKind::Other, "Unsupported Python version")),
+            },
+            PythonImplementation::PyPy => match self.version {
+                Some((3, 7)) => ("libpypy3-c.def", include_str!("libpypy3-c.def")),
+                Some((3, 8)) => ("libpypy3-c.def", include_str!("libpypy3-c.def")),
+                Some((3, 9)) => ("libpypy3.9-c.def", include_str!("libpypy3.9-c.def")),
+                _ => return Err(Error::new(ErrorKind::Other, "Unsupported PyPy version")),
+            },
         };
 
         let mut defpath = out_dir.to_owned();
@@ -309,19 +340,14 @@ impl DllToolCommand {
             return Ok(DllToolCommand::Zig { command, machine });
         }
 
-        match (arch, env) {
-            // 64-bit MinGW-w64 (aka `x86_64-pc-windows-gnu`)
-            ("x86_64", "gnu") => Ok(DllToolCommand::Mingw {
-                command: Command::new(DLLTOOL_GNU),
-            }),
-
-            // 32-bit MinGW-w64 (aka `i686-pc-windows-gnu`)
-            ("x86", "gnu") => Ok(DllToolCommand::Mingw {
-                command: Command::new(DLLTOOL_GNU_32),
+        match env {
+            // 64-bit and 32-bit MinGW-w64 (aka `{x86_64,i686}-pc-windows-gnu`)
+            "gnu" => Ok(DllToolCommand::Mingw {
+                command: get_mingw_dlltool(arch)?,
             }),
 
             // MSVC ABI (multiarch)
-            (_, "msvc") => {
+            "msvc" => {
                 if let Some(command) = find_lib_exe(arch) {
                     // MSVC tools use their own target architecture names...
                     let machine = match arch {
@@ -340,7 +366,7 @@ impl DllToolCommand {
                 }
             }
             _ => {
-                let msg = format!("Unsupported target arch '{}' or env ABI '{}'", arch, env);
+                let msg = format!("Unsupported target env ABI '{}'", env);
                 Err(Error::new(ErrorKind::Other, msg))
             }
         }
@@ -413,6 +439,31 @@ impl DllToolCommand {
     }
 }
 
+/// Chooses the appropriate MinGW-w64 `dlltool` executable
+/// for the target architecture.
+///
+/// Examines the user-provided `PYO3_MINGW_DLLTOOL` environment variable first
+/// and falls back to the default MinGW-w64 arch prefixes.
+fn get_mingw_dlltool(arch: &str) -> Result<Command> {
+    if let Ok(user_dlltool) = env::var("PYO3_MINGW_DLLTOOL") {
+        Ok(Command::new(&user_dlltool))
+    } else {
+        let prefix_dlltool = match arch {
+            // 64-bit MinGW-w64 (aka `x86_64-pc-windows-gnu`)
+            "x86_64" => Ok(DLLTOOL_GNU),
+            // 32-bit MinGW-w64 (aka `i686-pc-windows-gnu`)
+            "x86" => Ok(DLLTOOL_GNU_32),
+            // AArch64?
+            _ => {
+                let msg = format!("Unsupported MinGW target arch '{}'", arch);
+                Err(Error::new(ErrorKind::Other, msg))
+            }
+        }?;
+
+        Ok(Command::new(prefix_dlltool))
+    }
+}
+
 /// Finds the `zig` executable (when built by ``maturin --zig`).
 ///
 /// Examines the `ZIG_COMMAND` environment variable
@@ -477,6 +528,15 @@ mod tests {
                 .generate(&dir)
                 .unwrap();
         }
+
+        // PyPy
+        for minor in 7..=9 {
+            ImportLibraryGenerator::new("x86_64", "gnu")
+                .version(Some((3, minor)))
+                .implementation(PythonImplementation::PyPy)
+                .generate(&dir)
+                .unwrap();
+        }
     }
 
     #[cfg(unix)]
@@ -507,6 +567,15 @@ mod tests {
                 .generate(&dir)
                 .unwrap();
         }
+
+        // PyPy
+        for minor in 7..=9 {
+            ImportLibraryGenerator::new("x86_64", "msvc")
+                .version(Some((3, minor)))
+                .implementation(PythonImplementation::PyPy)
+                .generate(&dir)
+                .unwrap();
+        }
     }
 
     #[test]
